require("kara3d.vector")
require("kara3d.matrix")

-- ********** DEFINITION AND GENERIC FUNCTIONS **********

quaternion = {class = "quaternion"}
quaternion.__index = quaternion

-- Creates the quaternion ai + bj + dk + d
function quaternion.new(a, b, c, d)
  local q = {x = a or 0, y = b or 0, z = c or 0, w = d or 1}
  setmetatable(q, quaternion)
  return q
end

-- Clones this quaternion
function quaternion:clone()
  return quaternion.new(self.x, self.y, self.z, self.w)
end

-- Returns this quaternion's coordinates, separated by commas
function quaternion:tostring()
  return self.x .. ", " .. self.y .. ", " .. self.z .. ", " .. self.w
end

-- ********** STATIC FUNCTIONS **********

-- Returns the quaternion 0i + 0j + 0k + 1, which is no rotation
function quaternion.identity() return quaternion.new() end

-- Returns the product of two (normalized) quaternions
function quaternion.mul_quaternion(q1, q2)
  local b, c, d, a = q1:normalized():unpack()
  local f, g, h, e = q2:normalized():unpack()
  return quaternion.new(a*f + b*e + c*h - d*g,
                        a*g + c*e + d*f - b*h,
                        a*h + d*e + b*g - c*f,
                        a*e - b*f - c*g - d*h)
end

-- Returns a vector equal to the vector v (size >= 3)
-- rotated by the quaternion q
function quaternion.mul_vector(q, v)
  local v3 = vector.new(3, v.values)
  local u1 = vector.new(3, {q.x, q.y, q.z})
  local u2 = vector.new(3, {-q.x, -q.y, -q.z})
  local d = vector.dot(u1, v3)
  local c = vector.cross(u1, v3)
  return 2*d * u1 + (q.w*q.w - u1:length2()) * v3 + 2*q.w * c
end

-- Returns a quaternion representing the rotation from -> to (size = 3)
function quaternion.from_to_rotation(from, to)
  local v0 = from:normalized()
	local v1 = to:normalized()
	local d = v0:dot(v1)

  if (math.abs(d) < 0.001) then return quaternion.identity() end

  return quaternion.axis_angle(vector.cross(v0, v1):normalize(), math.acos(d))
end

-- Returns a quaternion representing the rotation of angle around axis
-- If angle is nil, then the norm of axis will be used as the angle
function quaternion.axis_angle(axis, angle)
  angle = angle or axis:length()

  local naxis = axis:normalized()
  local hcos = math.cos(0.5 * angle)
  local hsin = math.sin(0.5 * angle)

  return quaternion.new(naxis:x() * hsin, naxis:y() * hsin, naxis:z() * hsin, hcos)
end

-- ********** INSTANCE FUNCTIONS **********

-- Sets this quaternion's coordinates to these values
-- Arguments may be 4 numbers or 1 table with indexes x, y, z and w
function quaternion:set(x, y, z, w)
  if (type(x) == "table" and x.class == "quaternion") then
    self.x = x.x
    self.y = x.y
    self.z = x.z
    self.w = x.w
  else
    self.x = x or 0
    self.y = y or 0
    self.z = z or 0
    self.w = w or 1
  end
  return self
end

-- Returns the inverse of this quaternion
function quaternion:inv()
  return quaternion.new(-self.x, -self.y, -self.z, -self.w)
end

-- Returns the 4 coordinates of this quaternion
function quaternion:unpack()
  return self.x, self.y, self.z, self.w
end

-- Normalizes this quaternion, so that its norm is 1
function quaternion:normalize()
  local len = self:length()
  if len ~= 0 and len ~= 1 then
      self.x = self.x / len
      self.y = self.y / len
      self.z = self.z / len
      self.w = self.w / len
  end
  return self
end

-- Returns a quaternion which is equal to this quaternion, normalized
function quaternion:normalized()
  local q = self:clone()
  q:normalize()
  return q
end

-- Returns the square of the norm of this quaternion
function quaternion:length2()
  return self.x * self.x + self.y * self.y + self.z * self.z + self.w * self.w
end

-- Returns the norm of this quaternion
function quaternion:length()
  return math.sqrt(self:length2())
end

-- Returns the rotation matrix (4x4) corresponding to this quaternion
function quaternion:to_matrix()
  local x, y, z, w = self:unpack()
  local m = matrix.new(4, 4,
    {1 - 2*y*y - 2*z*z,     2*x*y - 2*w*z,     2*w*y + 2*x*z, 0,
         2*w*z + 2*x*y, 1 - 2*x*x - 2*z*z,     2*y*z - 2*w*x, 0,
         2*x*z - 2*w*y,     2*w*x + 2*y*z, 1 - 2*x*x - 2*y*y, 0,
                     0,                 0,                 0, 1})
  return m
end
