require("kara3d.vector")
require("kara3d.matrix")
require("kara3d.quaternion")
require("kara3d.transform")
require("kara3d.light")
require("kara3d.camera")
require("utils")
re = require("re")

-- ********** DEFINITION AND GENERIC FUNCTIONS **********

shape = {class = "shape"}
shape.__index = shape

face = {class = "face"}
face.__index = face

material = {class = "material"}
material.__index = material



function material.new(amb, diff, spec, sh, al)
  local m = {ambient = vector.new(3, amb) or vector.new(3, {1, 1, 1}), 
            diffuse = vector.new(3, diff) or vector.new(3, {1, 1, 1}),
            specular = vector.new(3, spec) or vector.new(3, {1, 1, 1}),
            shininess = sh or 1,
            alpha = al or 0}
  setmetatable(m, material)
  return m
end



-- Creates a new face with vertices c, normal n and tags t
function face.new(v, n, t)
  local f = {vertices = v or {}, normal = n or -vector.cardinal(3, 3), tags = t}
  setmetatable(f, face)
  return f
end



function face.shade(N, P, mat, lights, ambient, cam)
  local Ka = mat.ambient
  local Kd = mat.diffuse
  local Ks = mat.specular
  local shine = mat.shininess

  local V = (cam.transform:position(true) - P):normalized()

  local I = Ka * ambient

  for i = 1, #lights do
    local l = lights[i]
    local L = (l.transform:position(true) - P):normalized()
    local R = 2 * math.max(vector.dot(L, N), 0) * N - L

    -- Diffuse component
    I = I + Kd * math.max(vector.dot(L, N), 0) * l.diffuse

    -- Specular component
    I = I + Ks * math.pow(math.max(vector.dot(R, V), 0), shine) * l.specular
  end
  
  return I:clamp(0, 255)
end


-- Creates a new shape with transform t and faces f
function shape.new(t, f, mat)
  local s = {transform = t, faces = f or {}, material = mat or material.new()}
  setmetatable(s, shape)
  return s
end


-- Clones this shape
function shape:clone()
  return shape.new(self.transform, table.copy(self.faces))
end

-- ********** INSTANCE FUNCTIONS **********

function shape:set_material(mat)
  self.material = mat
  return self
end



-- Generates a line for each face of this shape, with the indicated tags
-- The boolean cull determines if the faces facing away from the screen are
-- drawn or not
function shape:draw(subs, line, tags, cull, cam, ambient, lights)
  local l = table.copy(line)
  local p = self.transform:position(true)
  local r = self.transform:rotation(true)
  local c = vector.cardinal(4, 4)
  local layer = line.layer

  local pv = cam and cam:matrix() or matrix.identity(4, 4)
  local cam_inv_rot = cam and cam.transform:rotation(true):inv() or quaternion.identity()
  local xres, yres, _, _ = aegisub.video_size()

  local m = pv * self.transform:matrix(true)

  for i = 1, #self.faces do
    local face = self.faces[i]
    local norm = quaternion.mul_vector(r, face.normal)
    local n = quaternion.mul_vector(cam_inv_rot, norm)

    if not (cull and n:z() > 0) then
      local vs = face.vertices

      local v = m * (vs[#vs] + c)
      v = (1 / v:w()) * v
      local sum = vector.zero(4)
      local str = (tags and "{" .. tags .. "}" or "")
        .. (face.tags and "{" .. face.tags .. "}" or "")
        .. "{\\an7\\pos(" .. (xres / 2) .. ", " .. (yres / 2) .. ")\\p1}"
        .. "m " .. v:x() .. " " .. v:y() .. " "

      for i = 1, #vs do
        v = m * (vs[i] + c)
        sum = sum + v
        v = (1 / v:w()) * v
        str = str .. "l " .. v:x() .. " " .. v:y() .. " "
      end

      local P = (p + sum) * (1/#vs)
      

      local color = face.shade(n, P, self.material, lights, ambient, cam)
      local colorstr = string.format("{\\c%s\\alpha%s}",
        ass_color(color:x(), color:y(), color:z()),
        ass_alpha(self.material.alpha)
      )

      l.text = colorstr .. str
      -- layers must be between 0 and 9999999999
      l.layer = 500000000 + math.floor((p + sum):z() / #vs)
      l.effect = "fx"
      subs.append(l)
    end
  end

  return ""
end

-- ********** BASIC SHAPE FUNCTIONS **********

-- Returns a cube shape of side length size, that follows the transform t
-- Schematic of the vertices:
--      1________5
--     /|      /|
--  2/__|____/6 |
--  |  3|_ _|_ _|7     __ X
--  |  /    |  /     /|
-- 4|/______|/8    Z  Y
function shape.cube(t, size)
  local hs = size / 2
  local vs = {vector.new(3, {-hs, -hs, -hs}), vector.new(3, {-hs, -hs,  hs}),
              vector.new(3, {-hs,  hs, -hs}), vector.new(3, {-hs,  hs,  hs}),
              vector.new(3, { hs, -hs, -hs}), vector.new(3, { hs, -hs,  hs}),
              vector.new(3, { hs,  hs, -hs}), vector.new(3, { hs,  hs,  hs})}

  local left =   face.new({vs[1], vs[2], vs[4], vs[3]}, -vector.cardinal(3, 1))
  local right =  face.new({vs[5], vs[6], vs[8], vs[7]},  vector.cardinal(3, 1))
  local bottom = face.new({vs[1], vs[2], vs[6], vs[5]}, -vector.cardinal(3, 2))
  local top =    face.new({vs[3], vs[4], vs[8], vs[7]},  vector.cardinal(3, 2))
  local back =   face.new({vs[1], vs[5], vs[7], vs[3]}, -vector.cardinal(3, 3))
  local front =  face.new({vs[2], vs[6], vs[8], vs[4]},  vector.cardinal(3, 3))

  return shape.new(t, {left, right, bottom, top, back, front})
end


--[[
  Load shape from mesh file. Only .off files supported for now, though
]]--
function shape.load(t, file)
  local vertices = {}
  local faces = {}
  local vertexNumber = nil
  local faceNumber = nil
  local vertexRead = 0
  local facesRead = 0
  local linenumber = 0

  for line in io.lines(file) do
    linenumber = linenumber + 1
    local blankMatch = re.match(line, "OFF|#.*")
    if blankMatch == nil then
      local matches = re.split(line, "[[:blank:]]+", true)
      if matches ~= nil then
        if vertexNumber == nil or faceNumber == nil then
            vertexNumber = tonumber(matches[1])
            faceNumber = tonumber(matches[2])
        else
          if vertexRead < vertexNumber then
              vertices[#vertices+1] = vector.new(3, {tonumber(matches[1]), tonumber(matches[2]), tonumber(matches[3])})
              vertexRead = vertexRead + 1
          else
            local faceSize = tonumber(matches[1])
            local faceVertices = {}
            for i = 1, faceSize do
                faceVertices[i] = vertices[tonumber(matches[i+1]) + 1]
            end
            local normal = vector.cross(faceVertices[2] - faceVertices[1],
                                        faceVertices[3] - faceVertices[1])
                                  :normalized()
            faces[#faces+1] = face.new(faceVertices, normal)
            facesRead = facesRead + 1
          end
        end
      end
    end
  end

  return shape.new(t, faces)
end
