-- ********** DEFINITION AND GENERIC FUNCTIONS **********

vector = {class = "vector"}
vector.__index = vector

-- Creates a vector of size n, initialized with values vals
-- Uninitialized values are set to 0
function vector.new(n, vals)
  if (type(n) == "table") then vals = n n = #n end
  vals = vals or {}
  local v = {size = n or 3, values = {}}
  for i = 1, v.size do
    v.values[i] = i <= #vals and vals[i] or 0
  end
  setmetatable(v, vector)
  return v
end

-- Clones this vector
function vector:clone() return vector.new(self.size, self.values) end

-- Returns this vector's coordinates, separated by commas
function vector:tostring()
  local str = ""
  for i = 1, self.size - 1 do
    str = str .. self:get(i) .. ", "
  end
  return str .. self:get(self.size)
end

-- ********** STATIC FUNCTIONS **********

-- Returns a vector full of 0s, with size size
function vector.zero(size) return vector.new(size) end

-- Returns a vector full of 1s, with size size
function vector.one(size)
  local vals = {}
  for i = 1, size do vals[i] = 1 end
  return vector.new(size, vals)
end

-- Returns a vector of size size, filled with 0s and a 1 at index
function vector.cardinal(size, index)
  local vals = {}
  for i = 1, size do vals[i] = i == index and 1 or 0 end
  return vector.new(size, vals)
end

-- Returns the dot product of two vectors
function vector.dot(a, b)
  local min_size = math.min(a.size, b.size)
  local sum = 0
  for i = 1, min_size do sum = sum + a.values[i] * b.values[i] end
  return sum
end

-- Returns the cross product of two vectors of size 3
function vector.cross(a, b)
  return vector.new(3, {a.values[2] * b.values[3] - a.values[3] * b.values[2],
                        a.values[3] * b.values[1] - a.values[1] * b.values[3],
                        a.values[1] * b.values[2] - a.values[2] * b.values[1]})
end

-- Returns a vector which is the projection of the vector from to the vector to
function vector.project_vector(from, to) return (vector.dot(from, to) / to:length2()) * to end

-- Returns a vector which is the projection of the vector from to the plane of normal normal
function vector.project_plane(from, normal) return from - vector.project_vector(from, normal) end

-- Returns the square of the distance between two vectors
function vector.distance2(a, b) return (a - b):length2() end

-- Returns the distance between two vectors
function vector.distance(a, b) return (a - b):length() end

-- Returns the angle between two vectors
function vector.angle(from, to) return math.acos(vector.dot(from, to) / (from:length() * to:length())) end

-- Returns a vector lerped between from and to, with parameter t
function vector.lerp(from, to, t) return t * from + (1 - t) * to end

-- ********** INSTANCE FUNCTIONS **********

-- Sets the ith cell of this vector to value
function vector:set(i, value) self.values[i] = value return self end

-- Returns the ith value of this vector
function vector:get(i) return self.values[i] end

-- Sets the size of this vector to n
-- Uninitialized values are set to 0, and overflowing values are set to nil
function vector:set_size(n)
  if (self.size == n) then return self end
  for i = n + 1, self.size do self:set(i, nil) end
  for i = self.size + 1, n do self:set(i, 0) end
  self.size = n
  return self
end

-- Sets the values of this vector to vals
-- Uninitialized values are set to 0
function vector:set_values(vals)
  vals = vals or {}
  for i = 1, #vals do self:set(i, vals[i]) end
  for i = #vals + 1, self.size do self:set(i, 0) end
  self.size = #vals
  return self
end

-- Returns the square of the norm of this vector
function vector:length2()
  local sum = 0
  for i = 1, self.size do sum = sum + self:get(i) * self:get(i) end
  return sum
end

-- Returns the norm of this vector
function vector:length() return math.sqrt(self:length2()) end

-- Returns ||v|| == 1, v being this vector
function vector:is_unit() return self:length2() == 1 end

-- Sets the coordinates of this vector so that it has the same direction,
-- but has a norm equal to 1
function vector:normalize()
  local len = self:length()
  if len == 0 or len == 1 then return self end
  for i = 1, self.size do self:set(i, self:get(i) / len) end
  return self
end

-- Returns a vector with the same direction as this vector, but with norm 1
function vector:normalized() return self:clone():normalize() end

-- Convenience functions that return the first four coordinates
function vector:x() return self.values[1] end
function vector:y() return self.values[2] end
function vector:z() return self.values[3] end
function vector:w() return self.values[4] end

-- Clamp vector values
function vector:clamp(min, max)
  for i = 1, self.size do self:set(i, clamp(self:get(i), min, max)) end 
  return self
end

-- ********** OPERATOR OVERLOADING **********

-- + binary operator overload
-- Returns a vector c such that for all i, c(i) = a(i) + b(i)
function vector.__add(a, b)
  local small = a.size < b.size and a or b
  local big = a.size < b.size and b or a
  local r = big:clone()
  for i = 1, small.size do r:set(i, r:get(i) + small:get(i)) end
  return r
end

-- - unary operator overload
-- Returns a vector c such that for all i, c(i) = -a(i)
function vector.__unm(a)
  local r = vector.new(a.size)
  for i = 1, r.size do r:set(i, -(a:get(i))) end
  return r
end

-- - binary operator overload
-- Returns a vector c such that for all i, c(i) = a(i) - b(i)
function vector.__sub(a, b)
  local oppb = -b
  return a + oppb
end

-- * binary operator overload
-- If one of the arguments is a scalar, scales the other argument by it
-- Else, returns a vector c such that for all i, c(i) = a(i) * b(i)
function vector.__mul(a, b)
  if type(a) == "number" and type(b) ~= "number" then
    local r = vector.new(b.size)
    for i = 1, r.size do r:set(i, a * b:get(i)) end
    return r
  elseif type(a) ~= "number" and type(b) == "number" then
    local r = vector.new(a.size)
    for i = 1, r.size do r:set(i, a:get(i) * b) end
    return r
  elseif type(a) ~= "number" and type(b) ~= "number" then
    local small = a.size < b.size and a or b
    local big = a.size < b.size and b or a
    local r = vector.new(big.size)
    for i = 1, small.size do r:set(i, small:get(i) * big:get(i)) end
    return r
  else
    return a * b
  end
end

-- == binary operator overload
-- Returns "for all i, a(i) == b(i)""
function vector.__eq(a, b)
  if (a.size ~= b.size) then return false end
  for i = 1, a.size do if (a:get(i) ~= b:get(i)) then return false end end
  return true
end
